/***********************************************************************
 *                                                                     *
 * gifconv.c - main routine for gifconv                                *
 *                                                                     *
 * Author: Tim Adye (T.J.Adye@rl.ac.uk), based heavily on XWPICK by    *
 *         Evgeni Chernyaev (chernaev@mx.ihep.su).                     *
 *         Includes contributions from                                 *
 *         Jean-Damien Durand (Jean-Damien.Durand@cern.ch)             *
 *                                                                     *
 * Copyright (C) 1993, 1994 by Evgeni Chernyaev.                       *
 *                                                                     *
 * Permission to use, copy, modify, and distribute this software and   *
 * its documentation for non-commercial purpose is hereby granted      *
 * without fee, provided that the above copyright notice appear in all *
 * copies and that both the copyright notice and this permission       *
 * notice appear in supporting documentation.                          *
 *                                                                     *
 * This software is provided "as is" without express or implied        *
 * warranty.                                                           *
 *                                                                     *
 ***********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef __STDC__
# define ARGS(alist) alist
# define CONST const
#else
# define ARGS(alist) ()
# define CONST
#endif

char ProgName[] = "gifconv";
char ProgVers[] = "Version 2.21b  Rev: 22/05/98 - by Tim Adye";
#define MAX_FILENAME 80     /* Maximum output (or prompted) filename length */
#define SCANF_FILENAME "%80s"   /* Uses MAX_FILENAME */
CONST char *FileName = NULL;
CONST char **InFile = NULL;
int nInFile = 0;
extern int verbose;
extern int showComment;
int xsize = 0;
int ysize = 0;

#ifndef VMS_STYLE
# ifdef __VMS
#   define VMS_STYLE   1
# else
# ifdef VMS
#   define VMS_STYLE   1
# else
#   define VMS_STYLE   0
# endif
# endif
#endif

#if VMS_STYLE
# define MAX_ARG     50
# define MAX_ARGLEN 256
#endif


typedef unsigned char byte;
typedef void (*fpl) ARGS((int,int,byte *));
typedef void (*fpb) ARGS((byte));
typedef char Image;     /* Should be void, but VAX C doesn't like that */

static Image   *image;
static FILE    *out = NULL;

   /*   G L O B A L   P R E D E F I N I T I O N S   */

long PS_LZWencode ARGS((int,int,int,int,CONST byte *,CONST byte *,CONST byte *,CONST byte *,fpl,fpb,int,int,int));
long PSencode     ARGS((int,int,int,byte *,byte *,byte *,byte *,fpl,fpb));
long GIFencode    ARGS((int,int,int,byte *,byte *,byte *,byte *,fpl,fpb));
long PCXencode    ARGS((int,int,int,byte *,byte *,byte *,byte *,fpl,fpb));
long PPMencode    ARGS((int,int,int,byte *,byte *,byte *,byte *,fpl,fpb));
long PICTencode   ARGS((int,int,int,byte *,byte *,byte *,byte *,fpl,fpb));
Image *GIFOpenImage  ARGS((CONST char *,int *,int *,int *,int *,int *,int *));
void   GIFCloseImage ARGS((Image *));
byte   GIFGetPixel   ARGS((CONST Image *,int,int));
void   GIFget_scline ARGS((CONST Image *,int,int,byte *));

   /*   L O C A L   P R E D E F I N I T I O N S   */

static int CheckName   ARGS((CONST char *));
static void get_scline ARGS((int,int,byte *));
static void put_byte   ARGS((byte));
static int usage       ARGS((void));
static int gifconv     ARGS((int,CONST char**));
#if VMS_STYLE
static int vms_parser  ARGS((int,char **, char*));
#endif

   /*   L O C A L   F U N C T I O N S   */

static int CheckName (format)
          CONST char *format;
{
  int ifrm;

  ifrm   = 0;
  if (format) {
    if (strncmp(format,"ps",  2)== 0 || strncmp(format,"PS",  2)== 0) ifrm = 1;
    if (strncmp(format,"eps", 3)== 0 || strncmp(format,"EPS", 3)== 0) ifrm = 2;
    if (strncmp(format,"epsi",4)== 0 || strncmp(format,"EPSI",4)== 0) ifrm = 3;
    if (strncmp(format,"gif", 3)== 0 || strncmp(format,"GIF", 3)== 0) ifrm = 4;
    if (strncmp(format,"pcx", 3)== 0 || strncmp(format,"PCX", 3)== 0) ifrm = 5;
    if (strncmp(format,"pict",3)== 0 || strncmp(format,"PICT",3)== 0) ifrm = 6;
    if (strncmp(format,"ppm", 3)== 0 || strncmp(format,"PPM", 3)== 0) ifrm = 7;
    if (strncmp(format,"psrl",3)== 0 || strncmp(format,"PSRL",3)== 0) ifrm = 8;
  }
  return ifrm;
}

static void get_scline(y, width, scline)
                   int y, width;
                  byte scline[];
{
  GIFget_scline (image, y, width, scline);
}

static void put_byte
#ifdef __STDC__
   (byte b)
#else
(b) byte b;
#endif
{
  if (ferror(out) == 0) fputc(b, out);
}

static int usage()
{
  fprintf(stderr,"%s: usage: %s [options...] [<gif-infiles> [<outfile>]]\n",
                 ProgName,ProgName);
  fprintf(stderr,"%s: stdout used if only one name specified, or outfile is -\n",ProgName);
  fprintf(stderr,"%s: possible options:\n",ProgName);
  fprintf(stderr,"  -reverse     transfer image to reverse colors\n");
  fprintf(stderr,"  -gray        transfer image to grayscale\n");
  fprintf(stderr,"  -norotate    don\'t rotate postscript image to fit\n");
  fprintf(stderr,"  -page        include page numbers in postscript\n");
  fprintf(stderr,"  -size sz     encapsulated postscript image size (eg. -size 200 or 100x200)\n");
  fprintf(stderr,"  -verbose     display extra GIF decoding information\n");
  fprintf(stderr,"  -format frm  set output format (default=ps for stdout)\n");
  fprintf(stderr,
 "%s: possible output file names: *.ps,*.eps,*.epsi,*.gif,*.pcx,*.pict,*.ppm\n",
    ProgName);
  return(1);
}

#if VMS_STYLE
static int vms_parser(argc,argv,strng)
                  int argc;
                 char *argv[], strng[];
{
  int i, j, k, n;
  char c;

  if (argc == 1) return 1;

  for (strng[0]='\0', k=i=1; i<argc; i++) {
    j = 0;
    do {
      c = argv[i][j++];
      switch (c) {
      case '/':
	if (strng[k-1] != '\0') strng[k++] = '\0';
	strng[k++] = '-';
	break;
      case '=': case '\0':
	if (strng[k-1] != '\0') strng[k++] = '\0';
        break;
      default:
	strng[k++] = c;
	break;
      }
    } while (c != '\0');
  }

  for (n=i=0; i<k; i++) {
    if (strng[i] == '\0') argv[++n] = &strng[i+1];
  }
  return n;
}
#endif /* VMS_STYLE==1 */

/***********************************************************************
 *                                                                     *
 * Name: gifconv                                     Date:    16.02.93 *
 * Author: E.Chernyaev (IHEP/Protvino)               Revised: 01.04.94 *
 *                                                            01.08.94 *
 *                                                            01.09.94 *
 * Revised:Tim Adye                                           26.08.94 *
 *                                                            15.05.96 *
 *                                                            22.05.98 *
 * Function: Converts a GIF file into one of the following:-           *
 *                                                                     *
 * FileName: '*.ps,*.eps,*epsi'   - PostScript                         *
 *           '*.gif'  - GIF-format  (Graphics Interchange Format)      *
 *           '*.pcx'  - PCX-format  (IBM PC)                           *
 *           '*.pict' - PICT-format (Macintosh)                        *
 *           '*.ppm'  - PPM-format  (see PBM Plus library)             *
 *                                                                     *
 ***********************************************************************/
static int gifconv
#ifdef __STDC__
            (int argc, CONST char **argv)
#else
(argc, argv) int argc; CONST char **argv;
#endif
{
  int           ifrever, ifgray, ifnorotate, ifform, ifstdout, ifpage;
  int           ifcolor;
  int           i, ifrm, maxcol;
  int           x, y, width, height, ncol, R[256], G[256], B[256];
  int           igif, ipnum;
  long          igray, irep;
  byte          r[256], g[256], b[256], *scline;
  CONST char    *oform;
  char          InFileString[MAX_FILENAME], OutFileString[MAX_FILENAME];
  CONST char    stdoutFlag[]= "-";

  /*   G E T   P A R A M E T E R S   */

  ifrever = ifgray = ifnorotate = ifpage = ifform = ifstdout = 0;
  verbose = showComment = 0;
  oform = NULL;
  InFile= (CONST char**) malloc( (argc-1) * sizeof(char*) );

  for (i=1; i<argc; i++) {
    if (argv[i][0] != '-' || argv[i][1] == '\0' )
      InFile[nInFile++]= argv[i];
    else if (strncmp(argv[i],"-reverse", 2)==0 || strncmp(argv[i],"-R",2)==0) {
      ifrever = 1;
    }
    else if (strncmp(argv[i],"-gray",    2)==0 || strncmp(argv[i],"-G",2)==0) {
      ifgray  = 1;
    }
    else if (strncmp(argv[i],"-norotate",2)==0 || strncmp(argv[i],"-N",2)==0) {
      ifnorotate = 1;
    }
    else if (strncmp(argv[i],"-page",    2)==0 || strncmp(argv[i],"-P",2)==0) {
      ifpage  = 1;
    }
    else if (strncmp(argv[i],"-verbose", 2)==0 || strncmp(argv[i],"-V",2)==0) {
      verbose = 1;
      showComment = 1;
    }
    else if (strncmp(argv[i],"-size",   2)==0 || strncmp(argv[i],"-S",2)==0) {
/*    We allow for -size 200x300, and the bounding box will be set to these
      values, but the PS header ensures that the aspect ratio remains 1:1. */
      if (++i < argc && argv[i][0] != '-') {
        char c = '\0';
        xsize = ysize = 0;
        sscanf (argv[i], "%dx%d%c", &xsize, &ysize, &c);
        if (xsize==0 || ysize==0 || c!='\0') {
          xsize = ysize = 0;
          c = 'x';      /* Allow 200x (same as 200) */
          sscanf (argv[i], "%d%c", &xsize, &c);
          if (xsize==0 || c!='x') {
            xsize = 0;
            c = '\0';
            sscanf (argv[i], "x%d%c", &ysize, &c);
            if (ysize==0 || c!='\0') {
              fprintf(stderr,"%s: invalid -size %s\n",ProgName,argv[i]);
              return(1);
            }
          }
        }
      }else{
        fprintf(stderr,"%s: missing size on -size\n",ProgName);
        return(1);
      }
    }
    else if (strncmp(argv[i],"-format", 2)==0 || strncmp(argv[i],"-F",2)==0) {
      ifform  = 1;
      if (++i < argc && argv[i][0] != '-') {
	oform = argv[i];
      }else{
        fprintf(stderr,"%s: missing format on -format\n",ProgName);
        return(1);
      }
    }
    else return usage();
  }

  switch (nInFile) {
  case 0:
    fprintf(stderr,"******************************************************\n");
    fprintf(stderr,"*                                                    *\n");
    fprintf(stderr,"*        GIFCONV 2.21b - Convert a GIF Image         *\n");
    fprintf(stderr,"*                                                    *\n");
    fprintf(stderr,"*  Input format: GIF (Graphics Interchange Format)   *\n");
    if (ifform == 0) {
    fprintf(stderr,"*                                                    *\n");
    fprintf(stderr,"*  Possible output file names:                       *\n");
    fprintf(stderr,"*                       *.ps   (PostScript)          *\n");
    fprintf(stderr,"*                       *.eps  (Encapsulated PS)     *\n");
    fprintf(stderr,"*                       *.epsi (EPS with preview)    *\n");
    fprintf(stderr,"*                       *.gif  (GIF)                 *\n");
    fprintf(stderr,"*                       *.pcx  (IBM PC)              *\n");
    fprintf(stderr,"*                       *.pict (Macintosh)           *\n");
    fprintf(stderr,"*                       *.ppm  (PBM+ library)        *\n");
    } else {
    fprintf(stderr,"*  Output format: %s\n", oform);
    }
    fprintf(stderr,"*                                                    *\n");
    fprintf(stderr,"******* Evgeni Chernyaev (chernaev@mx.ihep.su) *******\n");

    fprintf(stderr,"\nInput  File Name: ");
    scanf(SCANF_FILENAME,InFileString);
    InFile[0] = InFileString;
    nInFile = 1;

    fprintf(stderr,"\nOutput File name: ");
    scanf(SCANF_FILENAME,OutFileString);
    FileName = OutFileString;
    break;
  case 1:
    FileName = stdoutFlag;
    break;
  default:
    FileName = InFile[--nInFile];
  }
#ifdef DEBUG
  for (i= 0; i<nInFile; i++)
    fprintf(stderr,"%s ", InFile[i]);
  fprintf(stderr,"-> %s\n",FileName);
#endif
  ifstdout = (FileName[0] == '-' && FileName[1] == '\0');

  /*  S E T   O U T P U T   F O R M A T  */

  if (ifstdout != 0 && ifform == 0)
    ifrm = 1;    /* Default to postscript if writing to stdout */
  else {
    if (ifform == 0) {
      oform = strrchr(FileName,'.');
      if (oform) oform++;
    }
    ifrm= CheckName (oform);

    if (ifrm == 0) {
      if (ifform == 0) {
	fprintf(stderr,"%s: error in file name extention\n",ProgName);
	fprintf(stderr,
		"%s: possible file names: *.ps,*.eps,*.epsi,*.gif,*.pcx,*.pict,*.ppm\n",
		ProgName);
      }else{
	fprintf(stderr,"%s: error in format type\n",ProgName);
	fprintf(stderr,
		"%s: possible formats: ps, eps, epsi, gif, pcx, pict, ppm\n",
		ProgName);
      }
      return(1);
    }
  }

  /*  O P E N   F I L E  */

  if (ifstdout)
    out = stdout;
  else {
    out = fopen(FileName,"w+");
    if (!out) {
      fprintf(stderr,"%s: unable to open file: %s\n",ProgName,FileName);
      return(1);
    }
  }

  ipnum= -1;
  for (igif=0; igif<nInFile; igif++) {
    image= GIFOpenImage(InFile[igif], &width, &height, &ncol, R, G, B);

    /*   N O R M A L I S E   C O L O R S   */

    maxcol = 0;
    for (i=0; i<ncol; i++) {
      if (maxcol < R[i]) maxcol = R[i];
      if (maxcol < G[i]) maxcol = G[i];
      if (maxcol < B[i]) maxcol = B[i];
      r[i] = 0;
      g[i] = 0;
      b[i] = 0;
    }
    if (maxcol != 0) {
      for (i=0; i<ncol; i++) {
        r[i] = R[i] * 255 / maxcol;
        g[i] = G[i] * 255 / maxcol;
        b[i] = B[i] * 255 / maxcol;
      }
    }

    /*  I N V E R T   C O L O R S ,   I F   N E E D E D  */

    if (ifrever != 0) {
      maxcol = 0;
      for (i=0; i<ncol; i++) {
        r[i] = 255 - r[i];
        g[i] = 255 - g[i];
        b[i] = 255 - b[i];
        if (maxcol < r[i]) maxcol = r[i];
        if (maxcol < g[i]) maxcol = g[i];
        if (maxcol < b[i]) maxcol = b[i];
      }
      if (maxcol != 0) {
        for (i=0; i<ncol; i++) {
          r[i] = r[i] * 255 / maxcol;
          g[i] = g[i] * 255 / maxcol;
          b[i] = b[i] * 255 / maxcol;
        }
      }
    }

    /*   G R A Y S C A L E   ?   */

    if (ifgray == 0) {
      ifcolor = 0;
      for (i=0; i<ncol; i++) {
        if (r[i] != g[i] || r[i] != b[i]) ifcolor = 1;
      }
      ifgray = 1 - ifcolor;
    }

    /*  E N C O D E   I M A G E   */

    scline = (byte*) malloc (width * sizeof(byte));
    if (!scline) {
      fprintf(stderr,"Unable to allocate space for %d-byte scan line\n",width);
      return(1);
    }

    if (ifrm < 4) {
      if (ifpage) ipnum= igif+1;
      i    = ifnorotate*20 + ifgray*10 + ifrm - 1;
      irep = PS_LZWencode(i,width,height,ncol,r,g,b,scline,get_scline,put_byte,igif,nInFile,ipnum);
    }
    if (ifgray != 0) {
      for (i=0; i<ncol; i++) {
        igray = (r[i]*299L + g[i]*587L + b[i]*114L) / 1000L;
        r[i]  = igray;
        g[i]  = igray;
        b[i]  = igray;
      }
    }
    if (ifrm == 4)
      irep = GIFencode(width,height,ncol,r,g,b,scline,get_scline,put_byte);
    if (ifrm == 5)
      irep = PCXencode(width,height,ncol,r,g,b,scline,get_scline,put_byte);
    if (ifrm == 6)
      irep = PICTencode(width,height,ncol,r,g,b,scline,get_scline,put_byte);
    if (ifrm == 7)
      irep = PPMencode(width,height,ncol,r,g,b,scline,get_scline,put_byte);
    if (ifrm == 8)
      irep = PSencode(width,height,ncol,r,g,b,scline,get_scline,put_byte);

    free (scline);
    GIFCloseImage (image);

    if (ferror(out) == 0) {
      fprintf(stderr,"%s: %s %d colors %dx%d size %d bytes\n",
                    InFile[igif],FileName,ncol,width,height,irep);
    }else{
      fprintf(stderr,"%s: problems during file writing\n",ProgName);
      return(1);
    }
  }
  return(0);
}

int
main(ArgC, ArgV)
 int ArgC;
 char *ArgV[];
{
  int ret;
#if VMS_STYLE
  int i, argc;
  char *argv[MAX_ARG], mess[MAX_ARGLEN];

  argc = ArgC;
  for (i=0; i<argc; i++) argv[i] = ArgV[i];
  argc = vms_parser(argc,argv,mess);
  ret= gifconv(argc,argv);
#else
  ret= gifconv(ArgC,ArgV);
#endif

  if (out)    fclose(out);
  if (InFile) free(InFile);
  exit(ret);
}
